(* miniML abstract syntax *)

(* To simplify the implementation of typed closure conversion, we
   assume that in the miniML source program all the binders have been
   explicitely tagged with their type by the programmer (see the
   abstract syntax of Fun and Let).  We do not need to propagate this
   type informations to the target-language. *)

(* Identifiers are represented by strings.  The parser always returns
   a term where all binders are distinct. *)

(* The abstract type of miniML types *)
type exprtype =
  | Type_base of string
  | Type_var of string
  | Type_arrow of exprtype * exprtype
  | Type_product of exprtype * exprtype

(* The abstract type of miniML terms *)
type constant =
  | CInt of int
  | CBool of bool

type expression =
  | Var of string
  | Const of constant
  | Op of string
  | Fun of string * exprtype * expression
  | App of expression * expression
  | Pair of expression * expression
  | Fst of expression
  | Snd of expression
  | Let of string * exprtype * expression * expression

(* The abstract type of target-language types *)
type tl_exprtype =
  | TL_Type_base of string
  | TL_Type_var of string
  | TL_Type_fun of tl_exprtype * tl_exprtype * tl_exprtype
  | TL_Type_product of tl_exprtype * tl_exprtype
  | TL_Type_record of (string * tl_exprtype) list
  | TL_Type_exists of string * tl_exprtype

(* The abstract type of target-language terms *)
type tl_expression =
  | TL_Var of string
  | TL_Const of constant
  | TL_Op of string
  | TL_Fun of string * string * tl_expression
  | TL_Call of tl_expression * tl_expression * tl_expression
  | TL_Pair of tl_expression * tl_expression
  | TL_Fst of tl_expression
  | TL_Snd of tl_expression
  | TL_Let of string * tl_expression * tl_expression
  | TL_Record of (string * tl_expression) list
  | TL_Proj of tl_expression * string
  | TL_Pack of tl_exprtype * tl_expression
  | TL_Open of tl_expression * string * tl_expression






