(* Programming exercise no. 1 *)
open Ast

let rec substitute v_s x_s a_s =
  match a_s with
  | TL_Var s -> if s = x_s then v_s else a_s
  | TL_Const i -> a_s
  | TL_Op _ -> failwith "not implemented"
  | TL_Fun (x1,x2,a) -> 
      if (x_s = x1) || (x_s = x2) then a_s 
      else TL_Fun (x1, x2, substitute v_s x_s a)

  | TL_Let (x,a1,a2) -> 
      TL_Let (x, substitute v_s x_s a1, 
	      if x_s = x then a2 else substitute v_s x_s a2)
  | TL_Call (a1,a2,a3) -> 
      TL_Call (substitute v_s x_s a1, substitute v_s x_s a2, 
	       substitute v_s x_s a3)
  | TL_Pair (a1,a2) -> TL_Pair (substitute v_s x_s a1, substitute v_s x_s a2)
  | TL_Fst a -> TL_Fst (substitute v_s x_s a)
  | TL_Snd a -> TL_Snd (substitute v_s x_s a)
  | TL_Record l -> 
      let subst_item (e,a) = (e, substitute v_s x_s a) in
      TL_Record (List.map subst_item l)
  | TL_Proj (a,e) -> TL_Proj (substitute v_s x_s a, e)
  | TL_Pack (t,a) -> TL_Pack (t, substitute v_s x_s a)
  | TL_Open (a1,x,a2) -> 
      TL_Open (substitute v_s x_s a1, x,
	       if x_s = x then a2 else substitute v_s x_s a2)

let rec evaluate a = 
(*  print_endline (">> "^( Util.print_tl_expression a)); *)
  let result =
  match a with
  | TL_Var s -> failwith ("undefined variable "^s)
  | TL_Const i -> TL_Const i
  | TL_Op _ -> failwith "not implemented"
  | TL_Fun (x1,x2,a) -> TL_Fun (x1,x2,a)

  | TL_Call (a1,a2,a3) -> 
(*      print_endline ("CALL "^( Util.print_tl_expression a)); *) 
      let v1 = evaluate a1 in
(*      print_endline ("v1 = "^( Util.print_tl_value v1)); *)
      let v2 = evaluate a2 in
(*      print_endline ("v2 = "^( Util.print_tl_value v2));   *)
      let v3 = evaluate a3 in
(*      print_endline ("v3 = "^( Util.print_tl_value v3));   *)
      ( match v1 with
      | TL_Fun (x1,x2,a') -> 
(*     print_endline ("REC CALL "^( Util.print_tl_expression a'));  *)
          evaluate (substitute v3 x2 (substitute v2 x1 a'))
      | _ -> failwith "applying non-functional value" )

  | TL_Pair (a1, a2) -> TL_Pair (evaluate a1, evaluate a2)
  | TL_Fst a1 -> 
      ( match (evaluate a1) with
      | TL_Pair (v1,v2) -> v1
      | _ -> failwith "fst of non-pair value" )
  | TL_Snd a1 -> 
      ( match (evaluate a1) with
      | TL_Pair (v1,v2) -> v2
      | _ -> failwith "snd of non-pair value" )

  | TL_Let (x,a1,a2) -> evaluate (substitute (evaluate a1) x a2)

  | TL_Record l ->
      let evaluate_item (e,a) = (e, evaluate a) in
      TL_Record (List.map evaluate_item l)

  | TL_Proj (a,e) ->
      ( match (evaluate a) with
      | TL_Record l -> 
          ( try 
            List.assoc e l
          with Not_found -> failwith "undefined label" )
      | _ -> failwith "proj of non-record value" )

  | TL_Pack (t,a) -> TL_Pack (t, evaluate a)

  | TL_Open (a1,x,a2) ->
      ( match (evaluate a1) with
      | TL_Pack (t,v1) -> evaluate (substitute v1 x a2)
      | _ -> failwith "open of non-pack value" )
  in
(*  print_endline ("<< "^(Util.print_tl_value result)); *)
  result
