(* test *)

(* write your own test functions, and write your own set of examples *)

open Ast
open Util
open Progex1
open Progex2

(* as an example, we parse a string, pretty-print the parsed
   expression and type, perform typed closure conversion, and evaluate
   the resulting expression *)

let example s =
  print_endline "\n ************************** \n";
  let (a,tau) = parse_expression s in
  print_endline "\n*** Parsed expression:\n";
  print_endline(print_expression a);
  print_endline "\n*** Type of parsed expression:\n";
  print_endline(print_type tau);
  let (a',tau') = translate [] [] a tau in
  print_endline "\n*** Translated expression:\n";
  print_endline (print_tl_expression a');
  let v = evaluate a' in
  print_endline "\n*** Result of evaluation:\n";
  print_endline (print_tl_expression v);
  print_endline "\n ************************** \n"

let _ =
  (* 1- nothing to do *)
  example "let (x:bool) = true in (x,false) : bool * bool";
  (* 2- tcc with empty env *)
  example "(fun (x:int) -> (x,x)) 4 : (int * int)";
  (* 3- tcc with non-empty env *)
  example "(let (y:int)=5 in let (z:int)=6 in fun (x:int) -> (x,x)) 4 : int * int";
  (* 4- return a closure with empty env *)
  example "(fun (x:int) -> (x,x)) : int -> (int * int)";
  (* 5- return a closure with non-empty env *)
  example "let (y:int)=5 in let (z:int)=6 in fun (x:int) -> (x,x) : int -> (int * int)";
  (* 6- interleaving of tcc and env creation *)
  example "(fun (x:int) -> let (h:int) = 7 in fun (y:int) -> fun (z:int) -> (x,h,y,z)) 1 2 3 : int * (int * (int * int))";
  (* 7- higher-order *)
  example "(fun (x:int -> int) -> fun (z:int) -> x z) (fun (x:int) -> x) : (int -> int)";
  (* 8- higher-order and non empty env *)
  example "let (k:int)=4 in (fun (x:int -> int) -> fun (z:int) -> x k) (fun (x:int) -> x) 3 : int";

